import unittest
import asyncio
import os
import numpy as np
from pybela import Watcher, Streamer, Logger, Monitor

# all tests should be run with Bela connected and the bela-test project (in test/bela-test) running on the board


class test_Watcher(unittest.TestCase):

    def setUp(self):
        self.watcher = Watcher()
        self.watcher.connect()

    def tearDown(self):
        self.watcher.__del__()

    def test_list(self):
        self.assertEqual(len(self.watcher.list()["watchers"]), len(self.watcher.watcher_vars),
                         "Length of list should be equal to number of watcher variables")

    def test_start_stop(self):
        self.watcher.stop()
        self.assertTrue(self.watcher.ws_ctrl.close,
                        "Watcher ctrl websocket should be closed after stop")
        self.assertTrue(self.watcher.ws_data.close,
                        "Watcher data websocket should be closed after stop")


class test_Streamer(unittest.TestCase):

    def setUp(self):
        self.streamer = Streamer()
        self.streamer.connect()
        self.streaming_vars = [
            "myvar",  # dense double
            "myvar2",  # dense uint
            "myvar3",  # sparse uint
            "myvar4"  # sparse double
        ]
        self.saving_dir = "./test"

        self.saving_filename = "test_streamer_save.txt"

    def tearDown(self):
        self.streamer.__del__()

    def test_stream_n_values(self):
        n_values = 40

        streaming_buffer = self.streamer.stream_n_values(
            variables=self.streaming_vars[:2], n_values=n_values)

        # calc number of buffers needed to get n_values
        buffer_sizes = [
            self.streamer.get_data_length(var["type"], var["timestamp_mode"])
            for var in self.streamer.watcher_vars if var["name"] in self.streaming_vars[:2]]
        n_buffers = -(-n_values // min(buffer_sizes))

        # test data
        self.assertTrue(all(len(self.streamer.streaming_buffers_data[
                        var]) >= n_values for var in self.streaming_vars[:2]), "The streamed flat buffers for every variable should have at least n_values")
        self.assertTrue(all(len(streaming_buffer[
                        var]) == n_buffers for var in self.streaming_vars[:2]), "The streaming buffers queue should have at least n_values/buffer_size buffers for every variable")

    def __test_buffers(self, mode):
        # test data
        for var in [v for v in self.streamer.watcher_vars if v["name"] in self.streaming_vars]:

            # check buffers in streaming_buffers_queue have the right length
            self.assertTrue(all(len(_buffer["data"]) == var["data_length"] for _buffer in self.streamer.streaming_buffers_queue[var["name"]]),
                            f"The data buffers in self.streamer.streaming_buffers_queue should have a length of {var['data_length']} for a variable of type {var['type']} ")
            loaded = self.streamer.load_data_from_file(
                os.path.join(self.saving_dir, f"{var['name']}_{self.saving_filename}"))
            # check that the loader buffers have the right length
            self.assertTrue(all(len(_buffer["data"]) == var["data_length"] for _buffer in loaded),
                            "The loaded data buffers should have a length of {var['data_length']} for a variable of type {var['type']} ")
            # check that the number of buffers saved is the same as the number of buffers in self.streamer.streaming_buffers_queue
            self.assertTrue(len(self.streamer.streaming_buffers_queue[var['name']]) == len(loaded),
                            "The number of buffers saved should be equal to the number of buffers in self.streamer.streaming_buffers_queue (considering the queue length is long enough)")

        # check continuity of frames (only works for dense variables)
        if mode != "schedule":
            for var in [v for v in self.streamer.watcher_vars if v["name"] in self.streaming_vars[:2]]:
                for _buffer in self.streamer.streaming_buffers_queue[var["name"]]:
                    self.assertEqual(_buffer["ref_timestamp"], _buffer["data"][0],
                                     "The ref_timestamp and the first item of data buffer should be the same")
                    self.assertEqual(_buffer["ref_timestamp"]+var["data_length"]-1, _buffer["data"][-1],
                                     "The last data item should be equal to the ref_timestamp plus the length of the buffer")  # this test will fail if the Bela program has been streaming for too long and there are truncating errors. If this test fails, try stopping and rerunning hte Bela program again

        for var in self.streaming_vars:
            remove_file(os.path.join(self.saving_dir,
                        f"{var}_{self.saving_filename}"))

    def test_start_stop_streaming(self):
        async def async_test_start_stop():
            self.streamer.streaming_buffers_queue_length = 1000

            # delete any existing test files
            for var in self.streaming_vars:
                remove_file(os.path.join(self.saving_dir,
                                         f"{var}_{self.saving_filename}"))

            # stream with saving
            self.streamer.start_streaming(variables=self.streaming_vars,
                                          saving_enabled=True, saving_filename=self.saving_filename, saving_dir=self.saving_dir)
            # check streaming mode is FOREVER after start_streaming is called
            self.assertEqual(self.streamer._streaming_mode, "FOREVER",
                             "Streaming mode should be FOREVER after start_streaming")
            await asyncio.sleep(0.5)  # wait for some data to be streamed
            self.streamer.stop_streaming(variables=self.streaming_vars)
            # check streaming mode is OFF after stop_streaming

        asyncio.run(async_test_start_stop())
        self.assertEqual(self.streamer._streaming_mode, "OFF",
                         "Streaming mode should be OFF after stop_streaming")
        self.__test_buffers(mode="start_stop")

    def test_scheduling_streaming(self):
        async def async_test_scheduling_streaming():
            self.streamer.streaming_buffers_queue_length = 1000
            latest_timestamp = self.streamer.get_latest_timestamp()
            sample_rate = self.streamer.sample_rate
            timestamps = [latest_timestamp +
                          sample_rate] * len(self.streaming_vars)  # start streaming after ~1s
            durations = [sample_rate] * \
                len(self.streaming_vars)  # stream for 1s

            self.streamer.schedule_streaming(variables=self.streaming_vars,
                                             timestamps=timestamps,
                                             durations=durations,
                                             saving_enabled=True,
                                             saving_dir=self.saving_dir,
                                             saving_filename=self.saving_filename)

        asyncio.run(async_test_scheduling_streaming())
        self.__test_buffers(mode="schedule")


class test_Logger(unittest.TestCase):

    def setUp(self):
        self.logger = Logger()
        self.logger.connect()

        self.logging_vars = [
            "myvar",  # dense double
            "myvar2",  # dense uint
            "myvar3",  # sparse uint
            "myvar4"  # sparse double
        ]
        self.logging_dir = "./test"

    def tearDown(self):
        self.logger.__del__()

    def _test_logged_data(self, logger, logging_vars, local_paths):
        # common routine to test the data in the logged files
        data = {}
        for var in logging_vars:
            data[var] = logger.read_binary_file(
                file_path=local_paths[var], timestamp_mode=logger.get_prop_of_var(var, "timestamp_mode"))

            # test data
            timestamp_mode = logger.get_prop_of_var(var, "timestamp_mode")
            for _buffer in data[var]["buffers"]:
                self.assertEqual(_buffer["ref_timestamp"], _buffer["data"][0],
                                 "The ref_timestamp and the first item of data buffer should be the same")
                self.assertEqual(logger.get_prop_of_var(var, "data_length"), len(_buffer["data"]),
                                 "The length of the buffer should be equal to the data_length property of the variable")
                if _buffer["data"][-1] == 0:  # buffer has padding at the end
                    continue
                if timestamp_mode == "dense":
                    self.assertEqual(_buffer["ref_timestamp"]+logger.get_prop_of_var(var, "data_length")-1, _buffer["data"][-1],
                                     f"{var} {local_paths[var]} The last data item should be equal to the ref_timestamp plus the length of the buffer")
                elif timestamp_mode == "sparse":
                    inferred_timestamps = [_ + _buffer["ref_timestamp"]
                                           for _ in _buffer["rel_timestamps"]]
                    self.assertEqual(
                        inferred_timestamps, _buffer["data"], "The timestamps should be equal to the ref_timestamp plus the relative timestamps (sparse logging)")

    def test_logged_files_with_transfer(self):
        async def async_test_logged_files_with_transfer():

            # log with transfer
            file_paths = self.logger.start_logging(
                variables=self.logging_vars, transfer=True, logging_dir=self.logging_dir)
            await asyncio.sleep(0.5)
            self.logger.stop_logging()

            # test logged data
            self._test_logged_data(self.logger, self.logging_vars,
                                   file_paths["local_paths"])

            # clean local log files
            for var in file_paths["local_paths"]:
                remove_file(file_paths["local_paths"][var])
            # clean all remote log files in project
            self.logger.delete_all_bin_files_in_project()

        asyncio.run(async_test_logged_files_with_transfer())

    def test_logged_files_wo_transfer(self):
        async def async_test_logged_files_wo_transfer():

            # logging without transfer
            file_paths = self.logger.start_logging(
                variables=self.logging_vars, transfer=False, logging_dir=self.logging_dir)
            await asyncio.sleep(0.5)
            self.logger.stop_logging()

            # transfer files from bela
            local_paths = {}
            for var in file_paths["remote_paths"]:
                filename = os.path.basename(file_paths["remote_paths"][var])
                local_paths[var] = self.logger._generate_local_filename(
                    os.path.join(self.logging_dir, filename))
                self.logger.copy_file_from_bela(remote_path=file_paths["remote_paths"][var],
                                                local_path=local_paths[var])

            # test logged data
            self._test_logged_data(self.logger, self.logging_vars, local_paths)

            # clean log files
            for var in self.logging_vars:
                remove_file(local_paths[var])
                # self.logger.delete_file_from_bela(
                #     file_paths["remote_paths"][var])
            self.logger.delete_all_bin_files_in_project()

        asyncio.run(async_test_logged_files_wo_transfer())

    def test_scheduling_logging(self):
        async def async_test_scheduling_logging():
            latest_timestamp = self.logger.get_latest_timestamp()
            sample_rate = self.logger.sample_rate
            timestamps = [latest_timestamp +
                          sample_rate] * len(self.logging_vars)  # start logging after ~1s
            durations = [sample_rate] * len(self.logging_vars)  # log for 1s

            file_paths = self.logger.schedule_logging(variables=self.logging_vars,
                                                      timestamps=timestamps,
                                                      durations=durations,
                                                      transfer=True,
                                                      logging_dir=self.logging_dir)

            self._test_logged_data(self.logger, self.logging_vars,
                                   file_paths["local_paths"])

            # clean local log files
            for var in file_paths["local_paths"]:
                if os.path.exists(file_paths["local_paths"][var]):
                    os.remove(file_paths["local_paths"][var])
            self.logger.delete_all_bin_files_in_project()

            # # clean all remote log files in project
            # for var in file_paths["remote_paths"]:
            #     self.logger.delete_file_from_bela(
            #         file_paths["remote_paths"][var])

        asyncio.run(async_test_scheduling_logging())


class test_Monitor(unittest.TestCase):
    def setUp(self):
        self.monitor_vars = ["myvar", "myvar2", "myvar3", "myvar4"]
        self.period = 1000
        self.saving_filename = "test_monitor_save.txt"
        self.saving_dir = "./test"

        self.monitor = Monitor()
        self.monitor.connect()

    def tearDown(self):
        self.monitor.__del__()

    def test_peek(self):
        async def async_test_peek():
            peeked_values = self.monitor.peek()  # peeks at all variables by default
            for var in peeked_values:
                self.assertEqual(peeked_values[var]["timestamp"], peeked_values[var]["value"],
                                 "The timestamp of the peeked variable should be equal to the value")
        asyncio.run(async_test_peek())

    def test_period_monitor(self):
        async def async_test_period_monitor():
            self.monitor.start_monitoring(
                variables=self.monitor_vars[:2],
                periods=[self.period]*len(self.monitor_vars[:2]))
            await asyncio.sleep(0.5)
            monitored_values = self.monitor.stop_monitoring()

            for var in self.monitor_vars[:2]:  # assigned at every frame n
                self.assertTrue(np.all(np.diff(monitored_values[var]["timestamps"]) == self.period),
                                "The timestamps of the monitored variables should be spaced by the period")
                if var in ["myvar", "myvar2"]:  # assigned at each frame n
                    self.assertTrue(np.all(np.diff(monitored_values[var]["values"]) == self.period),
                                    "The values of the monitored variables should be spaced by the period")

        asyncio.run(async_test_period_monitor())

    def test_monitor_n_values(self):

        async def async_test_monitor_n_values():
            n_values = 25
            monitored_buffer = self.monitor.monitor_n_values(
                variables=self.monitor_vars[:2],
                periods=[self.period]*len(self.monitor_vars[:2]), n_values=n_values)

            for var in self.monitor_vars[:2]:
                self.assertTrue(np.all(np.diff(self.monitor.values[var]["timestamps"]) == self.period),
                                "The timestamps of the monitored variables should be spaced by the period")
                self.assertTrue(np.all(np.diff(self.monitor.values[var]["values"]) == self.period),
                                "The values of the monitored variables should be spaced by the period")
                self.assertTrue(all(len(self.monitor.streaming_buffers_data[
                    var]) >= n_values for var in self.monitor_vars[:2]), "The streamed flat buffers for every variable should have at least n_values")
                self.assertTrue(all(len(monitored_buffer[
                    var]["values"]) == n_values for var in self.monitor_vars[:2]), "The streaming buffers queue should have n_value for every variable")

        asyncio.run(async_test_monitor_n_values())

    def test_save_monitor(self):
        async def async_test_save_monitor():

            # delete any existing test files
            for var in self.monitor_vars:
                if os.path.exists(f"{var}_{self.saving_filename}"):
                    os.remove(f"{var}_{self.saving_filename}")

            self.monitor.start_monitoring(
                variables=self.monitor_vars,
                periods=[self.period]*len(self.monitor_vars),
                saving_enabled=True,
                saving_filename=self.saving_filename,
                saving_dir=self.saving_dir)
            await asyncio.sleep(0.5)
            monitored_buffers = self.monitor.stop_monitoring()

            for var in self.monitor_vars:
                loaded_buffers = self.monitor.load_data_from_file(os.path.join(self.saving_dir,
                                                                               f"{var}_{self.saving_filename}"))

                self.assertEqual(loaded_buffers["timestamps"], monitored_buffers[var]["timestamps"],
                                 "The timestamps of the loaded buffer should be equal to the timestamps of the monitored buffer")
                self.assertEqual(loaded_buffers["values"], monitored_buffers[var]["values"],
                                 "The values of the loaded buffer should be equal to the values of the monitored buffer")

            for var in self.monitor_vars:
                remove_file(os.path.join(self.saving_dir,
                                         f"{var}_{self.saving_filename}"))

        asyncio.run(async_test_save_monitor())


def remove_file(file_path):
    if os.path.exists(file_path):
        os.remove(file_path)


if __name__ == '__main__':
    if 0:
        unittest.main(verbosity=2)

    # select which tests to run
    n = 1
    for i in range(n):

        print(f"\n\n....Running test {i+1}/{n}")

        if 1:
            suite = unittest.TestSuite()
            if 1:
                suite.addTest(test_Watcher('test_list'))
                suite.addTest(test_Watcher('test_start_stop'))

            if 1:
                suite.addTest(test_Streamer('test_stream_n_values'))
                suite.addTest(test_Streamer('test_start_stop_streaming'))
                suite.addTest(test_Streamer('test_scheduling_streaming'))

            if 1:
                suite.addTest(test_Logger('test_logged_files_with_transfer'))
                suite.addTest(test_Logger('test_logged_files_wo_transfer'))
                suite.addTest(test_Logger('test_scheduling_logging'))

            if 1:
                suite.addTest(test_Monitor('test_peek'))
                suite.addTest(test_Monitor('test_period_monitor'))
                suite.addTest(test_Monitor('test_monitor_n_values'))
                suite.addTest(test_Monitor('test_save_monitor'))

            runner = unittest.TextTestRunner(verbosity=2)
            runner.run(suite)



import os
import asyncio
import aiofiles
import struct
import paramiko
from .Watcher import Watcher
from .utils import bcolors, print_error, print_info, print_ok, print_warning


class Logger(Watcher):
    def __init__(self, ip="192.168.7.2", port=5555, data_add="gui_data", control_add="gui_control"):
        """ Logger class

            Args:
                ip (str, optional): Remote address IP. If using internet over USB, the IP won't work, pass "bela.local". Defaults to "192.168.7.2".
                port (int, optional): Remote address port. Defaults to 5555.
                data_add (str, optional): Data endpoint. Defaults to "gui_data".
                control_add (str, optional): Control endpoint. Defaults to "gui_control".
        """
        super(Logger, self).__init__(ip, port, data_add, control_add)

        self._logging_mode = "OFF"
        self._logging_vars = []
        self._logging_transfer = True

        self.ssh_client = None
        self.sftp_client = None

        self._active_copying_tasks = []

        self._mode = "LOG"

    def start_logging(self, variables=[], transfer=True, logging_dir="./"):
        """ Starts logging session. The session can be ended by calling stop_logging().

        Args:
            variables (list of str, optional): List of variables to be logged. If no variables are passed, all variables in the watcher are logged. Defaults to [].
            transfer (bool, optional): If True, the logged files will be transferred automatically during the logging session. Defaults to True.

        Returns:
            list of str: List of local paths to the logged files.
        """

        remote_paths = self.__logging_common_routine(
            mode="FOREVER", timestamps=[], durations=[], variables=variables, logging_dir=logging_dir)

        local_paths = {}
        if transfer:
            for var in [v for v in self.watcher_vars if v["name"] in variables]:
                var = var["name"]
                local_path = os.path.join(
                    logging_dir, os.path.basename(remote_paths[var]))

                # if file already exists, throw a warning and add number at the end of the filename
                local_paths[var] = self._generate_local_filename(local_path)

                copying_task = self.copy_file_in_chunks(
                    remote_paths[var], local_paths[var])
                self._active_copying_tasks.append(copying_task)

        return {"local_paths": local_paths, "remote_paths": remote_paths}

    def schedule_logging(self, variables=[], timestamps=[], durations=[], transfer=True, logging_dir="./"):
        """Schedule logging session. The session starts at the specified timestamps and lasts for the specified durations. If the timestamp is in the past, the logging will start immediately. The session can be ended by calling stop_logging().

        Args:
            variables (list, optional): Variables to be logged. Defaults to [].
            timestamps (list, optional): Timestamps to start logging (one for each variable). Defaults to [].
            durations (list, optional): Durations to log for (one for each variable). Defaults to [].
            transfer (bool, optional): Transfer files to laptop automatically during logging session. Defaults to True.
            logging_dir (str, optional): Path to store the files. Defaults to "./".
        """
        async def _async_schedule_logging(variables, timestamps, durations, transfer, logging_dir):
            # checks types and if no variables are specified, stream all watcher variables (default)
            latest_timestamp = self.get_latest_timestamp()

            # check timestamps and duration types
            assert isinstance(
                timestamps, list) and all(isinstance(timestamp, int) for timestamp in timestamps), "Error: timestamps must be a list of ints."
            assert isinstance(
                durations, list) and all(isinstance(duration, int) for duration in durations), "Error: durations must be a list of ints."

            remote_paths = self.__logging_common_routine(
                mode="SCHEDULED", timestamps=timestamps, durations=durations, variables=variables, logging_dir=logging_dir)

            local_paths = {}
            if transfer:
                async def _async_check_if_file_exists_and_start_copying(var, timestamp):

                    diff_stamps = timestamp - latest_timestamp
                    while True:
                        _has_file_been_created = 0

                        remote_file_size = self.sftp_client.stat(
                            remote_paths[var]).st_size

                        if remote_file_size > 0:  # white till first buffers are written into the file
                            _has_file_been_created = 1
                            print_info(
                                f"Logging started for {var}...")
                            break  # Break the loop if the remote file size is non-zero

                        if _has_file_been_created:
                            break

                        # Wait before checking again
                        await asyncio.sleep(diff_stamps//(2*self.sample_rate))

                    # when file has been created
                    local_path = os.path.join(
                        logging_dir, os.path.basename(remote_paths[var]))

                    # if file already exists, throw a warning and add number at the end of the filename
                    local_paths[var] = self._generate_local_filename(
                        local_path)

                    copying_task = self.copy_file_in_chunks(
                        remote_paths[var], local_paths[var])
                    self._active_copying_tasks.append(copying_task)

                _active_checking_tasks = []
                for idx, var in enumerate(variables):
                    check_task = asyncio.create_task(
                        _async_check_if_file_exists_and_start_copying(var, timestamps[idx]))
                    _active_checking_tasks.append(check_task)

                # wait for the longest duration
                await asyncio.sleep(max(durations)//(self.sample_rate))
                self._logging_mode = "OFF"

                # wait for all the files to be created
                await asyncio.gather(*_active_checking_tasks, return_exceptions=True)
                # wait for all the files to be copied
                await asyncio.gather(*self._active_copying_tasks, return_exceptions=True)
                self._active_copying_tasks.clear()
                _active_checking_tasks.clear()
                self.sftp_client.close()

                # async version (non blocking)
                # async def _async_cleanup():
                #     await asyncio.gather(*self._active_copying_tasks, return_exceptions=True)
                #     self._active_copying_tasks.clear()
                #     self.sftp_client.close()
                # asyncio.run(_async_cleanup())

            return {"local_paths": local_paths, "remote_paths": remote_paths}

        return asyncio.run(_async_schedule_logging(variables=variables, timestamps=timestamps, durations=durations, transfer=transfer, logging_dir=logging_dir))

    def __logging_common_routine(self, mode, timestamps=[], durations=[], variables=[], logging_dir="./"):
        # checks types and if no variables are specified, stream all watcher variables (default)
        variables = self._var_arg_checker(variables)

        if not os.path.exists(logging_dir):
            os.makedirs(logging_dir)

        if self.is_logging():
            self.stop_logging()

        # self.start()  # start websocket connection -- done with .connect()
        self.connect_ssh()  # start ssh connection

        self._logging_mode = mode

        async def _async_send_logging_cmd_and_wait_for_response(var):
            # the logger responds with the name of the file where the variable is being logged in Bela
            # the logger responds with one message per variable -- so to keep track of responses it is easier to ask for a variable at a time rather than all at once
            self.send_ctrl_msg(
                {"watcher": [{"cmd": "log", "timestamps": timestamps, "durations": durations, "watchers": [var]}]})
            await self._log_response_available.wait()
            self._log_response_available.clear()
            return self._log_response

        remote_files = {}
        remote_paths = {}
        for var in variables:
            remote_files[var] = asyncio.run(_async_send_logging_cmd_and_wait_for_response(var))[
                "logFileName"]
            remote_paths[var] = f'/root/Bela/projects/{self.project_name}/{remote_files[var]}'

        return remote_paths

    def stop_logging(self, variables=[]):
        """ Stops logging session.

        Args:
            variables (list of str, optional): List of variables to stop logging. If none is passed, logging is stopped for all variables in the watcher. Defaults to [].
        """
        async def async_stop_logging(variables=[]):
            self._logging_mode = "OFF"
            if variables == []:
                # if no variables specified, stop streaming all watcher variables (default)
                variables = [var["name"] for var in self.watcher_vars]

            self.send_ctrl_msg(
                {"watcher": [{"cmd": "unlog", "watchers": variables}]})

            await asyncio.gather(*self._active_copying_tasks, return_exceptions=True)
            self._active_copying_tasks.clear()

            self.sftp_client.close()

            # self.stop()

        return asyncio.run(async_stop_logging(variables))

    def connect_ssh(self):
        """ Connects to Bela via ssh to transfer log files.
        """

        if self.sftp_client is not None:
            self.disconnect_ssh()

        self.ssh_client = paramiko.SSHClient()
        self.ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())

        # Workaround for no authentication:
        # https://github.com/paramiko/paramiko/issues/890#issuecomment-906893725
        try:
            self.ssh_client.connect(
                self.ip, port=22, username="root", password=None)
        except paramiko.SSHException as e:
            self.ssh_client.get_transport().auth_none("root")
        except Exception as e:
            print_error(
                f"Error while connecting to Bela via ssh: {e} {bcolors.ENDC}")
            return

        self.sftp_client = self.ssh_client.open_sftp()  # TODO handle exceptions better

    def disconnect_ssh(self):
        """ Disconnects from Bela via ssh.
        """
        if self.sftp_client:
            self.sftp_client.close()
        if self.ssh_client:
            self.ssh_client.close()

    def is_logging(self):
        """ Returns True if the logger is currently logging, false otherwise.

        Returns:
            bool: Logger status
        """
        return True if self._logging_mode != "OFF" else False

    def copy_file_in_chunks(self, remote_path, local_path,  chunk_size=2**12):
        """ Copies a file from the remote path to the local path in chunks. This function is called by start_logging() if transfer=True.

        Args:
            remote_path (str): Path to the file in Bela.
            local_path (str): Path to the file in the local machine (where the file is copied to)
            chunk_size (int, optional): Chunk size. Defaults to 2**12.

        Returns:
            asyncio.Task: Task that copies the file in chunks.
        """

        async def async_copy_file_in_chunks(remote_path, local_path, chunk_size=2**12):

            while True:
                # Wait for a second before checking again
                await asyncio.sleep(1)

                try:
                    remote_file = self.sftp_client.open(remote_path, 'rb')
                    remote_file_size = self.sftp_client.stat(
                        remote_path).st_size

                    if remote_file_size > 0:  # white till first buffers are written into the file
                        break  # Break the loop if the remote file size is non-zero

                except FileNotFoundError:
                    print_error(
                        f"Remote file '{remote_path}' does not exist.")
                    return None

            try:
                async with aiofiles.open(local_path, 'wb') as local_file:
                    while True:
                        chunk = remote_file.read(chunk_size)
                        # keep checking file whilst logging is still going on (in case a variable fills the buffers slowly)
                        if not chunk and self._logging_mode == "OFF":
                            break
                        await local_file.write(chunk)
                        print_ok(
                            f"\rTransferring {remote_path}-->{local_path}...", end="", flush=True)
                    await asyncio.sleep(0.1)  # flushed data
                    chunk = remote_file.read()
                    if chunk:
                        await local_file.write(chunk)
                    remote_file.close()
                    print_ok("Done.")

            except Exception as e:
                print_error(
                    f"Error while transferring file: {e}.")
                return None

            finally:
                await self._async_remove_item_from_list(self._active_copying_tasks, asyncio.current_task())

        return asyncio.create_task(async_copy_file_in_chunks(remote_path, local_path, chunk_size))

    def read_binary_file(self, file_path, timestamp_mode):
        """ Reads a binary file generated by the logger and returns a dictionary with the file contents.

        Args:
            file_path (str): Path of the file to be read.
            timestamp_mode (str): Timestamp mode of the variable. Can be "dense" or "sparse". 

        Returns:
            dict: Dictionary with the file contents.
        """

        file_size = os.path.getsize(file_path)
        assert file_size != 0, f"Error: The size of {file_path} is 0."

        def _parse_null_terminated_string(file):
            result = ""
            while True:
                char = file.read(1).decode('utf-8')
                if char == '\0':
                    break
                result += char
            return result

        with open(file_path, "rb") as file:

            # parse header
            name = _parse_null_terminated_string(file)
            var_name = _parse_null_terminated_string(file)
            _type = _parse_null_terminated_string(file)
            pid = struct.unpack("I", file.read(struct.calcsize("I")))[0]
            pid_id = struct.unpack("I", file.read(struct.calcsize("I")))[0]

            # if header size is not a multiple of 4, we need to add padding
            header_size = len(name) + len(var_name) + len(_type) + \
                3 + struct.calcsize("I") + struct.calcsize("I")
            if header_size % 4 != 0:
                file.read(4 - header_size % 4)  # padding

            _parse_type = 'i' if _type == 'j' else _type  # struct does not understand 'j'

            # parse data buffers
            parsed_buffers = []
            while True:
                # Read file buffer by buffer
                try:
                    data = file.read(self.get_buffer_size(
                        _parse_type, timestamp_mode))
                    if len(data) == 0:
                        break  # No more data to read
                    _parsed_buffer = self._parse_binary_data(
                        data, timestamp_mode, _parse_type)
                    parsed_buffers.append(_parsed_buffer)

                except struct.error as e:
                    print_error(str(e))
                    break  # No more data to read

        return {
            "project_name": name,
            "var_name": var_name,
            "type": _type,
            # "pid": pid,
            # "pid_id": pid_id,
            "buffers": parsed_buffers
        }

    # - utils

    # -- ssh copy utils

    def copy_file_from_bela(self, remote_path, local_path, verbose=True):
        """Copy a file from Bela onto the local machine.

        Args:
            remote_path (str): Path to the remote file to be copied.
            local_path (str): Path to the local file (where the file is copied to)
            verbose (bool, optional): Show info messages. Defaults to True.
        """
        self.connect_ssh()
        asyncio.run(self._async_copy_file_from_bela(
            remote_path, local_path, verbose))
        self.disconnect_ssh()

    def copy_all_bin_files_in_project(self, dir="./", verbose=True):
        """ Copies all .bin files in the specified remote directory using SFTP.

        Args:
            dir (str, optional): Path to the local directory where the files are copied to. Defaults to "./".
            verbose (bool, optional): Show info messages. Defaults to True.
        """
        remote_path = f'/root/Bela/projects/{self.project_name}'
        try:
            self.connect_ssh()
            copy_tasks = self._action_on_all_bin_files_in_project(
                "copy", dir)

            # wait until all files are copied
            asyncio.run(asyncio.gather(
                *copy_tasks, return_exceptions=True))

            if verbose:
                print_ok(
                    f"All .bin files in {remote_path} have been copied to {dir}.")
        except Exception as e:
            print_error(
                f"Error copying .bin files in {remote_path}: {e}")
        finally:
            self.disconnect_ssh()

    async def _async_copy_file_from_bela(self, remote_path, local_path, verbose=False):
        """ Copies a file from the remote path in Bela to the local path. This can be used any time to copy files from Bela to the host. 

        Args:
            remote_path (str): Path to the file in Bela.
            local_path (str): Path to the file in the local machine (where the file is copied to)
        """
        try:
            if os.path.exists(local_path):
                local_path = self._generate_local_filename(local_path)
            transferred_event = asyncio.Event()
            def callback(transferred, to_transfer): return transferred_event.set(
            ) if transferred == to_transfer else None
            self.sftp_client.get(remote_path, local_path, callback=callback)
            await asyncio.wait_for(transferred_event.wait(), timeout=3)
            if verbose:
                print_ok(
                    f"\rTransferring {remote_path}-->{local_path}... Done.")
            return transferred_event.is_set()
        except asyncio.TimeoutError:
            print_error("Timeout while transferring file.")
            return False  # File copy did not complete within the timeout
        except Exception as e:
            print_error(f"Error while transferring file: {e}")
            return False

    # -- ssh delete utils

    def delete_file_from_bela(self, remote_path, verbose=True):
        """Deletes a file from the remote path in Bela.

        Args:
            remote_path (str): Path to the remote file to be deleted. 
        """
        self.connect_ssh()
        asyncio.run(self._async_delete_file_from_bela(remote_path, verbose))
        self.disconnect_ssh()

    def delete_all_bin_files_in_project(self, verbose=True):
        """ Deletes all .bin files in the specified remote directory using SFTP.
        """
        remote_path = f'/root/Bela/projects/{self.project_name}'
        try:
            self.connect_ssh()
            deletion_tasks = self._action_on_all_bin_files_in_project(
                "delete")

            # wait until all files are deleted
            asyncio.run(asyncio.gather(
                *deletion_tasks, return_exceptions=True))

            if verbose:
                print_ok(
                    f"All .bin files in {remote_path} have been removed.")
        except Exception as e:
            print_error(
                f"Error deleting .bin files in {remote_path}: {e}")
        finally:
            self.disconnect_ssh()

    async def _async_delete_file_from_bela(self, remote_path, verbose=True):
        # this function doesn't return until the file has been deleted
        try:
            self.sftp_client.stat(remote_path)  # check if file exists
        except FileNotFoundError:
            print_error(
                f"Error: Remote file '{remote_path}' does not exist.")
            return

        while True:
            await asyncio.sleep(0.1)  # Adjust the interval as needed
            try:
                # Attempt to remove the file
                self.sftp_client.remove(remote_path)
            except FileNotFoundError:
                # File does not exist, it has been successfully removed
                if verbose:
                    print_ok(
                        f"File '{remote_path}' has been removed from Bela.")
                break
            except Exception as e:
                print_error(
                    f"Error while deleting file in Bela: {e} ")
                break

    # -- bunk task utils

    def _action_on_all_bin_files_in_project(self, action, local_dir=None):
        # List all files in the remote directory
        remote_path = f'/root/Bela/projects/{self.project_name}'
        file_list = self.sftp_client.listdir(remote_path)
        if len(file_list) == 0:
            print_warning(f"No .bin files in {remote_path}.")
            return

        # Iterate through the files and delete .bin files
        tasks = []
        for file_name in file_list:
            if file_name.endswith('.bin'):
                remote_file_path = f"{remote_path}/{file_name}"
                if action == "delete":
                    task = asyncio.create_task(
                        self._async_delete_file_from_bela(remote_file_path))
                elif action == "copy":
                    local_filename = os.path.join(local_dir, file_name)
                    task = asyncio.create_task(
                        self._async_copy_file_from_bela(remote_file_path, local_filename))
                else:
                    raise ValueError(f"Invalid action: {action}")
                tasks.append(task)

        return tasks

    def __del__(self):
        self.disconnect_ssh()  # disconnect ssh
        self.stop()  # stop websockets



from .Streamer import Streamer
import asyncio


class Monitor(Streamer):
    def __init__(self, ip="192.168.7.2", port=5555, data_add="gui_data", control_add="gui_control"):
        """ Monitor class

            Args:
                ip (str, optional): Remote address IP. If using internet over USB, the IP won't work, pass "bela.local". Defaults to "192.168.7.2".
                port (int, optional): Remote address port. Defaults to 5555.
                data_add (str, optional): Data endpoint. Defaults to "gui_data".
                control_add (str, optional): Control endpoint. Defaults to "gui_control".
        """

        super(Monitor, self).__init__(ip, port, data_add, control_add)

        self._mode = "MONITOR"

    def connect(self):
        super().connect()
        # longer queue for monitor since each buffer has only one value
        self.streaming_buffers_queue_length = 2000

    @property
    def values(self):
        """ Get monitored values from last monitoring session

        Returns:
            dict of dicts of list: Dict containing the monitored buffers for each variable in the watcher
        """
        values = {}
        for var in self.streaming_buffers_queue:
            values[var] = {"timestamps": [], "values": []}
            for _buffer in self.streaming_buffers_queue[var]:
                values[var]["timestamps"].append(_buffer["timestamp"])
                values[var]["values"].append(_buffer["value"])
        return values

    def peek(self, variables=[]):
        """ Peek at variables

            Args:
                variables (list, optional): List of variables to peek at. If no variables are specified, stream all watcher variables (default).

            Returns:
                dict: Dictionary of variables with their values
        """

        async def _async_peek(variables):
            # checks types and if no variables are specified, stream all watcher variables (default)
            variables = self._var_arg_checker(variables)
            self._peek_response = {var: None for var in variables}
            self.start_monitoring(variables, [1]*len(variables))
            await self._peek_response_available.wait()
            self._peek_response_available.clear()
            peeked_values = self._peek_response
            # set _peek_response again to None so that peek is not notified every time a new buffer is received (see Streamer._process_data_msg)
            self._peek_response = None
            self.stop_monitoring(variables)

            return peeked_values

        return asyncio.run(_async_peek(variables))

        # using list
        # res = self.list()
        # return {var: next(r["value"] for r in res if r["name"] == var) for var in variables}

    def start_monitoring(self, variables=[], periods=[], saving_enabled=False, saving_filename="monitor.txt", saving_dir="/."):
        """
        Starts the monitoring session. The session can be stopped with stop_monitoring().

        If no variables are specified, all watcher variables are monitored. If saving_enabled is True, the monitored data is saved to a local file. If saving_filename is None, the default filename is used with the variable name appended to its start. The filename is automatically incremented if it already exists. 

        Args:
            variables (list, optional): List of variables to be streamed. Defaults to [].
            periods (list, optional): List of monitoring periods. Defaults to [].
            saving_enabled (bool, optional): Enables/disables saving monitored data to local file. Defaults to False.
            saving_filename (str, optional) Filename for saving the monitored data. Defaults to None.
            saving_dir (str, optional): Directory for saving the monitored data. Defaults to "/.".
        """

        variables = self._var_arg_checker(variables)
        periods = self._check_periods(periods, variables)

        self.start_streaming(
            variables=variables, periods=periods, saving_enabled=saving_enabled, saving_filename=saving_filename, saving_dir=saving_dir)

    def monitor_n_values(self, variables=[], periods=[], n_values=1000, saving_enabled=False, saving_filename="monitor.txt"):
        """
        Monitors a given number of values. Since the data comes in buffers of a predefined size, always an extra number of frames will be monitored (unless the number of frames is a multiple of the buffer size). 

        Note: This function will block the main thread until n_values have been monitored. Since the monitored values come in blocks, the actual number of returned frames monitored may be higher than n_values, unless n_values is a multiple of the block size (monitor._streaming_block_size).

        To avoid blocking, use the async version of this function:
            monitor_task = asyncio.create_task(monitor.async_monitor_n_values(variables, n_values, periods, saving_enabled, saving_filename))
        and retrieve the monitored buffer using:
             monitored_buffers_queue = await monitor_task

        Args:
            variables (list, optional): List of variables to be monitored. Defaults to [].
            periods (list, optional): List of monitoring periods. Defaults to [].
            n_values (int, optional): Number of values to monitor for each variable. Defaults to 1000.
            delay (int, optional): _description_. Defaults to 0.
            saving_enabled (bool, optional): Enables/disables saving monitored data to local file. Defaults to False.
            saving_filename (_type_, optional) Filename for saving the monitored data. Defaults to None.

        Returns:
            monitored_buffers_queue (dict): Dict containing the monitored buffers for each streamed variable.
        """
        variables = self._var_arg_checker(variables)
        periods = self._check_periods(periods, variables)
        self.stream_n_values(variables, periods, n_values,
                             saving_enabled, saving_filename)
        return self.values

    async def async_monitor_n_values(self, variables=[], periods=[], n_values=1000, saving_enabled=False, saving_filename=None):
        """ 
        Asynchronous version of monitor_n_values(). Usage: 
            monitor_task = asyncio.create_task(monitor.async_monitor_n_values(variables, periods, n_values, saving_enabled, saving_filename)) 
        and retrieve the monitored buffer using:
            monitoring_buffers_queue = await monitor_task


        Args:
            variables (list, optional): List of variables to be monitored. Defaults to [].
            periods (list, optional): List of monitoring period. Defaults to [].
            n_values (int, optional): Number of values to monitor for each variable. Defaults to 1000.
            saving_enabled (bool, optional): Enables/disables saving monitored data to local file. Defaults to False.
            saving_filename (_type_, optional) Filename for saving the monitored data. Defaults to None.

        Returns:
            deque: Monitored buffers queue
        """
        variables = self._var_arg_checker(variables)
        periods = self._check_periods(periods, variables)

        self.async_stream_n_values(variables, periods, n_values,
                                   saving_enabled, saving_filename)

    def stop_monitoring(self, variables=[]):
        """
        Stops the current monitoring session for the given variables. If no variables are passed, the monitoring of all variables is interrupted.

        Args:
            variables (list, optional): List of variables to stop monitoring. Defaults to [].

        Returns:
            dict of dicts of lists: Dict containing the monitored buffers for each monitored variable
        """
        self.stop_streaming(variables)
        self._monitored_vars = None  # reset monitored variables
        # return only nonempty variables
        return {var: self.values[var] for var in self.values if self.values[var]["timestamps"] != []}

    def load_data_from_file(self, filename, flatten=True):
        """
        Loads data from a file saved through the saving_enabled function in start_monitoring() or monitor_n_values(). The file should contain a list of dicts, each dict containing a variable name and a list of values. The list of dicts should be separated by newlines.
        Args:
            filename (str): Filename
            flatten (bool, optional): If True, the returned list of values is flattened. Defaults to True.

        Returns:
            dict of lists: Dict with "timestamps" and "values" list if flatten is True, otherwise a list of dicts with "timestamp" and "value" keys.
        """
        loaded_buffers = super().load_data_from_file(filename)
        if flatten:
            flatten_loaded_buffers = {"timestamps": [], "values": []}
            for _buffer in loaded_buffers:
                flatten_loaded_buffers["timestamps"].append(
                    _buffer["timestamp"])
                flatten_loaded_buffers["values"].append(_buffer["value"])
            return flatten_loaded_buffers
        else:
            return loaded_buffers



import json
import copy
import os
import glob
import asyncio
import aiofiles  # async file i/o
from collections import deque  # circular buffers
from itertools import cycle
import warnings
import re

import bokeh.plotting
import bokeh.io
import bokeh.driving
from bokeh.resources import INLINE
from .Watcher import Watcher
from .utils import print_info, print_error, print_warning


class Streamer(Watcher):
    def __init__(self, ip="192.168.7.2", port=5555, data_add="gui_data", control_add="gui_control"):
        """ Streamer class

            Args:
                ip (str, optional): Remote address IP. If using internet over USB, the IP won't work, pass "bela.local". Defaults to "192.168.7.2".
                port (int, optional): Remote address port. Defaults to 5555.
                data_add (str, optional): Data endpoint. Defaults to "gui_data".
                control_add (str, optional): Control endpoint. Defaults to "gui_control".
        """

        super(Streamer, self).__init__(ip, port, data_add, control_add)

        # stores the list of monitored variables for each monitored session. cleaned after each monitoring session. used to avoid calling list() every time a new message is parsed
        self._monitored_vars = None

        # number of streaming buffers (not of data points!)
        self._streaming_buffers_queue_length = 1000
        self._streaming_buffers_queue = None
        self.last_streamed_buffer = {}

        self._streaming_mode = "OFF"  # OFF, FOREVER, N_VALUES, PEEK :: this flag prevents writing into the streaming buffer unless requested by the user using the start/stop_streaming() functions
        self._streaming_buffer_available = asyncio.Event()

        self._saving_enabled = False
        self._saving_filename = None
        self._saving_task = None
        self._active_saving_tasks = []
        self._saving_file_locks = {}

        self._peek_response_available = asyncio.Event()
        self._peek_response = None

        self._mode = "STREAM"

    # --- public methods --- #

    # - setters & getters

    @property
    def monitored_vars(self):
        """ Returns a list of monitored variables. If no variables are monitored, returns an empty list.

        Returns:
            list: list of monitored variables
        """
        if self._monitored_vars is None:  # avoids calling list() every time a new message is parsed
            _list = self.list()
            self._monitored_vars = self._filtered_watcher_vars(
                _list["watchers"], lambda var: var["monitor"])
            if self._monitored_vars == []:
                self._monitored_vars = None
        return self._monitored_vars

    @property
    def streaming_buffers_queue_length(self):
        """
        Returns:
            int: maximum number of streaming buffers allowed in self.streaming_buffers_queue
        """
        return self._streaming_buffers_queue_length

    @streaming_buffers_queue_length.setter
    def streaming_buffers_queue_length(self, value):
        """Sets the maximum number of streaming buffers allowed in self.streaming_buffers_queue. Warning: setting the streaming buffer value will result in deleting the current streaming buffer queue.
        """
        # TODO resize in terms of number of datapoints instead of number of buffers?
        self._streaming_buffers_queue_length = value
        self._streaming_buffers_queue = {var["name"]: deque(
            maxlen=self._streaming_buffers_queue_length) for var in self.watcher_vars}  # resize streaming buffer

    @property
    def streaming_buffers_queue(self):
        """Returns a dict where each key corresponds to a variable and each item to the variable's buffer queue. The queue has maximum length determined by streamer.streaming_buffers_queue_length. Each item of the queue is a received buffer of the form {"frame: int, "data": {"var1": [val1, val2, ...], "var2": [val1, val2, ...], ...} }

        Returns:
            dict: streaming buffers queue
        """
        # returns a dict of lists instead of a dict of dequeues
        return {key: list(value) for key, value in self._streaming_buffers_queue.items()}

    def start(self):
        """Starts the websocket connection and initialises the streaming buffers queue.
        """
        # self.connect()
        if not self.is_connected():
            print_warning(
                f'{"Monitor" if self._mode=="MONITOR" else "Streamer" } is not connected to Bela. Run {"monitor" if self._mode=="MONITOR" else "streamer"}.connect() first.')
            return 0
        self._streaming_buffers_queue = {var["name"]: deque(
            maxlen=self._streaming_buffers_queue_length) for var in self.watcher_vars}
        self.last_streamed_buffer = {
            var["name"]: {"data": [], "timestamps": []} for var in self.watcher_vars}

    @property
    def streaming_buffers_data(self):
        """Returns a dict where each key corresponds to a variable and each value to a flat list of the streamed values. Does not return timestamps of each datapoint since that depends on how often the variables are reassigned in the Bela code.
        Returns:
            dict: Dict of flat lists of streamed values.
        """
        data = {}
        for var in self.streaming_buffers_queue:
            data[var] = []
            for _buffer in self.streaming_buffers_queue[var]:
                data[var].extend(_buffer["data"] if self._mode !=
                                 "MONITOR" else [_buffer["value"]])
        return data

    # - streaming methods

    def __streaming_common_routine(self, variables=[], saving_enabled=False, saving_filename="var_stream.txt", saving_dir="./"):

        if self.is_streaming():
            print_warning("Stopping previous streaming session...")
            self.stop_streaming()  # stop any previous streaming

        if self.start() == 0:  # bela is not connected
            return

        if not os.path.exists(saving_dir):
            os.makedirs(saving_dir)

        self._saving_enabled = True if saving_enabled else False
        self._saving_filename = self._generate_filename(
            saving_filename, saving_dir) if saving_enabled else None

        # checks types and if no variables are specified, stream all watcher variables (default)
        variables = self._var_arg_checker(variables)

    def start_streaming(self, variables=[], periods=[], saving_enabled=False, saving_filename="var_stream.txt", saving_dir="./"):
        """
        Starts the streaming session. The session can be stopped with stop_streaming().

        If no variables are specified, all watcher variables are streamed. If saving_enabled is True, the streamed data is saved to a local file. If saving_filename is None, the default filename is used with the variable name appended to its start. The filename is automatically incremented if it already exists. 

        Args:
            variables (list, optional): List of variables to be streamed. Defaults to [].
            periods (list, optional): List of streaming periods. Streaming periods are used by the monitor and will be ignored if in streaming mode. Defaults to [].
            saving_enabled (bool, optional): Enables/disables saving streamed data to local file. Defaults to False.
            saving_filename (str, optional) Filename for saving the streamed data. Defaults to None.
        """

        self.__streaming_common_routine(
            variables, saving_enabled, saving_filename, saving_dir)

        self._streaming_mode = "FOREVER" if self._peek_response is None else "PEEK"
        if self._mode == "STREAM":
            if periods != []:
                warnings.warn(
                    "Periods list is ignored in streaming mode STREAM")
            self.send_ctrl_msg(
                {"watcher": [{"cmd": "watch", "watchers": variables}]})
            print_info(
                f"Started streaming variables {variables}... Run stop_streaming() to stop streaming.")
        elif self._mode == "MONITOR":
            periods = self._check_periods(periods, variables)
            self.send_ctrl_msg(
                {"watcher": [{"cmd": "monitor", "watchers": variables, "periods": periods}]})
            if self._streaming_mode == "FOREVER":
                print_info(
                    f"Started monitoring variables {variables}... Run stop_monitoring() to stop monitoring.")
            elif self._streaming_mode == "PEEK":
                print_info(f"Peeking at variables {variables}...")

    def stop_streaming(self, variables=[]):
        """
        Stops the current streaming session for the given variables. If no variables are passed, the streaming of all variables is interrupted.

        Args:
            variables (list, optional): List of variables to stop streaming. Defaults to [].

        Returns:
            streaming_buffers_queue (dict): Dict containing the streaming buffers for each streamed variable.
        """
        async def async_stop_streaming(variables=[]):
            # self.stop()

            _previous_streaming_mode = copy.copy(self._streaming_mode)

            self._streaming_mode = "OFF"
            if self._saving_enabled:
                self._saving_enabled = False
                self._saving_filename = None
                # await all active saving tasks
                await asyncio.gather(*self._active_saving_tasks, return_exceptions=True)
                self._active_saving_tasks.clear()

            if variables == []:
                # if no variables specified, stop streaming all watcher variables (default)
                variables = [var["name"] for var in self.watcher_vars]

            if self._mode == "STREAM" and _previous_streaming_mode != "SCHEDULE":
                self.send_ctrl_msg(
                    {"watcher": [{"cmd": "unwatch", "watchers": variables}]})
                print_info(f"Stopped streaming variables {variables}...")
            elif self._mode == "MONITOR" and _previous_streaming_mode != "SCHEDULE":
                self.send_ctrl_msg(
                    {"watcher": [{"cmd": "monitor", "periods": [0]*len(variables),  "watchers": variables}]})  # setting period to 0 disables monitoring
                if not _previous_streaming_mode == "PEEK":
                    print_info(f"Stopped monitoring variables {variables}...")

        return asyncio.run(async_stop_streaming(variables))

    def schedule_streaming(self, variables=[], timestamps=[], durations=[], saving_enabled=False, saving_filename="var_stream.txt", saving_dir="./"):
        """Schedule streaming of variables. The streaming session can be stopped with stop_streaming().

        Args:

            variables (list, optional): List of variables to be streamed. Defaults to [].
            timestamps (list, optional): Timestamps to start streaming (one for each variable). Defaults to [].
            durations (list, optional): Durations to stream for (one for each variable). Defaults to [].
            saving_enabled (bool, optional): Enables/disables saving streamed data to local file. Defaults to False.
            saving_filename (str, optional) Filename for saving the streamed data. Defaults to None.
            saving_dir (str, optional): Directory for saving the streamed data files. Defaults to "./".
        """

        self.__streaming_common_routine(
            variables, saving_enabled, saving_filename, saving_dir)

        self._streaming_mode = "SCHEDULE"

        self.send_ctrl_msg(
            {"watcher": [{"cmd": "watch", "timestamps": timestamps, "durations": durations, "watchers": variables}]})

        async def async_check_if_variables_have_been_streamed_and_stop():
            # poll to see when variables start streaming and when they stop
            started_streaming_vars = []
            finished_streaming_vars = []

            while not all(var in finished_streaming_vars for var in variables):

                for var in [v["name"] for v in self.watched_vars]:
                    if var not in started_streaming_vars:
                        started_streaming_vars.append(var)
                        print_info(f"Started streaming {var}...")

                for var in started_streaming_vars:
                    if var not in [v["name"] for v in self.watched_vars]:
                        finished_streaming_vars.append(var)
                        print_info(f"Stopped streaming {var}")

                await asyncio.sleep(0.1)

            self.stop_streaming()

        asyncio.run(
            async_check_if_variables_have_been_streamed_and_stop())

    def stream_n_values(self, variables=[], periods=[], n_values=1000, saving_enabled=False, saving_filename=None, saving_dir="./"):
        """
        Streams a given number of values. Since the data comes in buffers of a predefined size, always an extra number of frames will be streamed (unless the number of frames is a multiple of the buffer size). 

        Note: This function will block the main thread until n_values have been streamed. Since the streamed values come in blocks, the actual number of returned frames streamed may be higher than n_values, unless n_values is a multiple of the block size (streamer._streaming_block_size).

        To avoid blocking, use the async version of this function:
            stream_task = asyncio.create_task(streamer.async_stream_n_values(variables, n_values, periods, saving_enabled, saving_filename))
        and retrieve the streaming buffer using:
             streaming_buffers_queue = await stream_task

        Args:
            variables (list, optional): List of variables to be streamed. Defaults to [].
            periods (list, optional): List of streaming periods. Streaming periods are used by the monitor and will be ignored if in streaming mode. Defaults to [].
            n_values (int, optional): Number of values to stream for each variable. Defaults to 1000.
            delay (int, optional): _description_. Defaults to 0.
            saving_enabled (bool, optional): Enables/disables saving streamed data to local file. Defaults to False.
            saving_filename (str, optional) Filename for saving the streamed data. Defaults to None.
            saving_dir (str, optional): Directory for saving the streamed data. Defaults to "./".

        Returns:
            streaming_buffers_queue (dict): Dict containing the streaming buffers for each streamed variable.
        """
        return asyncio.run(self.async_stream_n_values(variables, periods, n_values, saving_enabled, saving_filename, saving_dir))

    async def async_stream_n_values(self, variables=[], periods=[], n_values=1000, saving_enabled=False, saving_filename="var_stream.txt", saving_dir="./"):
        """ 
        Asynchronous version of stream_n_values(). Usage: 
            stream_task = asyncio.create_task(streamer.async_stream_n_values(variables, n_values, saving_enabled, saving_filename)) 
        and retrieve the streaming buffer using:
            streaming_buffers_queue = await stream_task


        Args:
            variables (list, optional): List of variables to be streamed. Defaults to [].
            periods (list, optional): List of streaming periods. Streaming periods are used by the monitor and will be ignored if in streaming mode. Defaults to [].
            n_values (int, optional): Number of values to stream for each variable. Defaults to 1000.
            saving_enabled (bool, optional): Enables/disables saving streamed data to local file. Defaults to False.
            saving_filename (str, optional) Filename for saving the streamed data. Defaults to None.
            saving_dir (str, optional): Directory for saving the streamed data. Defaults to "./".

        Returns:
            deque: Streaming buffers queue
        """
        # resizes the streaming buffer size to n_values and returns it when full

        self.__streaming_common_routine(
            variables=variables, saving_enabled=saving_enabled, saving_filename=saving_filename, saving_dir=saving_dir)

        self._streaming_mode = "N_VALUES"  # flag cleared in __rec_msg_callback

        if self._mode == "STREAM":
            # if mode stream, each buffer has m values and we need to calc the min buffers needed to supply n_values

            # variables might have different buffer sizes -- the code below finds the minimum number of buffers needed to stream n_values for all variables
            buffer_sizes = [
                self.get_data_length(var["type"], var["timestamp_mode"])
                for var in self.watcher_vars if var["name"] in variables]

            # TODO add a warning when there's different buffer sizes ?

            # ceiling division
            n_buffers = -(-n_values // min(buffer_sizes))
            # using setter to automatically resize buffer
            self.streaming_buffers_queue_length = n_buffers

            if periods != []:
                warnings.warn(
                    "Periods list is ignored in streaming mode STREAM")
            self.send_ctrl_msg(
                {"watcher": [{"cmd": "unwatch", "watchers": [var["name"] for var in self.watcher_vars]}, {"cmd": "watch", "watchers": variables}]})
            print_info(
                f"Streaming {n_values} values for variables {variables}...")

        elif self._mode == "MONITOR":
            # if mode monitor, each buffer has 1 value so the length of the streaming buffer queue is equal to n_values
            self.streaming_buffers_queue_length = n_values

            periods = self._check_periods(periods, variables)
            self.send_ctrl_msg(
                {"watcher": [{"cmd": "monitor", "watchers": variables, "periods": periods}]})
            print_info(
                f"Monitoring {n_values} values for variables {variables} with periods {periods}...")

        # await until streaming buffer is full
        await self._streaming_buffer_available.wait()
        self._streaming_buffer_available.clear()

        # turns off listener, unwatches variables
        self.stop_streaming(variables)
        if self._mode == "MONITOR":
            self._monitored_vars = None  # reset monitored vars

        return self.streaming_buffers_queue

    # - utils

    def is_streaming(self):
        """Returns True if the streamer is currently streaming, False otherwise.

        Returns:
            bool: Streaming status bool
        """
        return True if self._streaming_mode != "OFF" else False

    def load_data_from_file(self, filename):
        """
        Loads data from a file saved through the saving_enabled function in start_streaming() or stream_n_values(). The file should contain a list of dicts, each dict containing a variable name and a list of values. The list of dicts should be separated by newlines.
        Args:
            filename (str): Filename

        Returns:
            list: List of values loaded from file
        """
        try:
            data = []
            with open(filename, "r") as f:
                while True:
                    line = f.readline()
                    if not line:
                        break
                    try:
                        data.append(json.loads(line))
                    except EOFError:  # reached end of file
                        break
        except Exception as e:
            print_error(f"Error while loading data from file: {e}")
            return None

        return data

    # - plotting

    def _bokeh_plot_data_app(self,
                             data,
                             x_var,
                             y_vars,
                             y_range=None,
                             rollover=None,
                             plot_update_delay=90):
        """Return a function defining a Bokeh app for streaming. The app is called in plot_data().
        Args:
            data (dict): Dict containing the data to be plotted. The dict should have the following structure: {"var1": {"data": [val1, val2, ...], "timestamps": [ts1, ts2, ...]}, "var2": {"data": [val1, val2, ...], "timestamps": [ts1, ts2, ...]}, ...}
            x_var (str): Variable to be plotted on the x axis
            y_vars (list): List of variables to be plotted on the y axis
            y_range (tuple, optional): Tuple containing the y axis range. Defaults to None. If none is given, the y axis range is automatically resized to fit the data.
            rollover (int, optional): Number of data points to keep on the plot. Defaults to None.
            plot_update_delay (int, optional): Delay between plot updates in ms. Defaults to 90.
        """
        # TODO add variable checkers

        def _app(doc):
            # Instantiate figures
            p = bokeh.plotting.figure(
                frame_width=500,
                frame_height=175,
                x_axis_label="timestamps",
                y_axis_label="value",
            )

            if y_range is not None:
                p.y_range = bokeh.models.Range1d(y_range[0], y_range[1])

            # No padding on x_range makes data flush with end of plot
            p.x_range.range_padding = 0

            # Create a dictionary to store ColumnDataSource instances for each y_var
            template = {"timestamps": [], **{var: [] for var in data}}
            source = bokeh.models.ColumnDataSource(template)

            # # Create line glyphs for each y_var
            colors = cycle([
                "#1f77b4", "#ff7f0e", "#2ca02c", "#d62728",
                "#9467bd", "#8c564b", "#e377c2", "#7f7f7f",
                "#bcbd22", "#17becf", "#1a55FF", "#FF1A1A"
            ])
            for y_var in y_vars:
                p.line(source=source, x="timestamps",
                       y=y_var, line_color=next(colors), legend_label=y_var)

            @bokeh.driving.linear()
            def update(step):
                # Update plot by streaming in data
                new_data = {"timestamps": [
                    data[x_var]["timestamp"]]if "timestamp" in data[x_var] else data[x_var]["timestamps"]}
                for y_var in y_vars:
                    new_data[y_var] = data[y_var]["data"] if isinstance(
                        data[y_var]["data"], list) else [data[y_var]["data"]]
                source.stream(new_data, rollover)

            doc.add_root(p)
            doc.add_periodic_callback(update, plot_update_delay)
        return _app

    def plot_data(self, x_var, y_vars, y_range=None, plot_update_delay=100, rollover=500):
        """ Plots a bokeh figure with the streamed data. The plot is updated every plot_update_delay ms. The plot is interactive and can be zoomed in/out, panned, etc. The plot is shown in the notebook.

        Args:
            x_var (str): Variable to be plotted on the x axis
            y_vars (list of str): List of variables to be plotted on the y axis
            y_range (float, float):  Tuple containing the y axis range. Defaults to None. If none is given, the y axis range is automatically resized to fit the data.
            plot_update_delay (int, optional): Delay between plot updates in ms. Defaults to 100.
            rollover (int, optional): Number of data points to keep on the plot. Defaults to 1000.
        """

        if self._mode == "MONITOR":
            raise NotImplementedError(
                "Plotting is not yet supported in monitor mode.")

        # check that x_var and y_vars are either streamed or monitored
        for _var in [x_var, *y_vars]:
            if not (_var in [var["name"] for var in self.watched_vars] or _var in [var["name"] for var in self.monitored_vars]):
                print_error(
                    f"PlottingError: {_var} is not being streamed or monitored.")
                return

        # check buffer lengths are the same
        # wait until streaming buffers have been populated
        async def wait_for_streaming_buffers_to_arrive():
            while not all(data['data'] for data in {
                    var: _buffer for var, _buffer in self.last_streamed_buffer.items() if var in y_vars}.values()):
                await asyncio.sleep(0.1)
        asyncio.run(wait_for_streaming_buffers_to_arrive())
        if len(y_vars) > 1 and not all([len(self.last_streamed_buffer[y_var]) == len(self.last_streamed_buffer[y_vars[0]]) for y_var in y_vars[1:]]):
            print_error(
                "PlottingError: plotting buffers of different length is not supported yet. Try using the same timestamp mode and type for your variables...")
            return

        bokeh.io.output_notebook(INLINE)
        bokeh.io.show(self._bokeh_plot_data_app(data={
            var: _buffer for var, _buffer in self.last_streamed_buffer.items() if var in y_vars}, x_var=x_var,
            y_vars=y_vars, y_range=y_range, plot_update_delay=plot_update_delay, rollover=rollover))

    # --- private methods --- #

    def _process_data_msg(self, msg):
        """ Process data message received from Bela. This function is called by the websocket listener when a data message is received.

        Args:
            msg (bytestring): Data message received from Bela
        """
        global _channel, _type
        try:
            _, __ = _channel, _type
        except NameError:  # initialise global variables to None
            _channel = None
            _type = None

        # in case buffer is received whilst streaming mode is on but parsed after streaming_enabled has changed
        _saving_enabled = copy.copy(self._saving_enabled)
        if self._streaming_mode != "OFF":
            if len(msg) in [3, 4]:  # _channel can be either 1 or 2 bytes long
                # parse buffer header
                _channel, _type = re.search(
                    r'(\d+).*?(\w)', msg.decode()).groups()
                _channel = int(_channel)

                assert _type in ['i', 'f', 'j', 'd',
                                 'c'], f"Unsupported type: {_type}"

                assert _type == self._watcher_vars[_channel][
                    'type'], f"Type mismatch: {_type} != {self._watcher_vars[_channel]['type']}"

                # convert unsigned int to int -- struct does not support unsigned ints
                _type = 'i' if _type == 'j' else _type

            elif len(msg) > 3 and _channel is not None and _type is not None:
                var_name = self._watcher_vars[_channel]['name']
                var_timestamp_mode = self._watcher_vars[_channel]["timestamp_mode"]

                # parse buffer body
                parsed_buffer = self._parse_binary_data(
                    msg, var_timestamp_mode, _type).copy()

                # fixes bug where data is shifted by period
                _var_streaming_buffers_queue = copy.deepcopy(
                    self._streaming_buffers_queue[var_name])
                _var_streaming_buffers_queue.append(parsed_buffer)
                self._streaming_buffers_queue[var_name] = _var_streaming_buffers_queue

                # populate last streamed buffer
                if self._mode == "STREAM":
                    self.last_streamed_buffer[var_name]["data"] = parsed_buffer["data"]
                    if var_timestamp_mode == "dense":
                        self.last_streamed_buffer[var_name]["timestamps"] = [
                            parsed_buffer["ref_timestamp"] + i for i in range(0, len(parsed_buffer["data"]))]
                    elif var_timestamp_mode == "sparse":  # sparse
                        self.last_streamed_buffer[var_name]["timestamps"] = [
                            parsed_buffer["ref_timestamp"] + i for i in parsed_buffer["rel_timestamps"]]
                elif self._mode == "MONITOR":
                    self.last_streamed_buffer[var_name] = {
                        "timestamp": parsed_buffer["timestamp"], "value": parsed_buffer["value"]}

                # save data to file if saving is enabled
                if _saving_enabled:
                    _saving_var_filename = os.path.join(os.path.dirname(
                        self._saving_filename), f"{var_name}_{os.path.basename(self._saving_filename)}")
                    # save the data asynchronously
                    saving_task = asyncio.create_task(
                        self._save_data_to_file(_saving_var_filename, parsed_buffer))
                    self._active_saving_tasks.append(saving_task)

                # response to .peek() call
                if self._mode == "MONITOR" and self._peek_response is not None:
                    # check that all the watched variables have been received
                    self._peek_response[self._watcher_vars[_channel]["name"]] = {
                        "timestamp": parsed_buffer["timestamp"], "value": parsed_buffer["value"]}
                    # notify peek() that data is available
                    if all(value is not None for value in self._peek_response.values()):
                        self._peek_response_available.set()

                # if streaming buffers queue is full for watched variables and streaming mode is n_values
                if self._streaming_mode == "N_VALUES":
                    obs_vars = self.watched_vars if self._mode == "STREAM" else self.monitored_vars
                    if all(len(self._streaming_buffers_queue[var["name"]]) == self._streaming_buffers_queue_length
                            for var in obs_vars):
                        self._streaming_mode = "OFF"
                        self._streaming_buffer_available.set()

    async def _save_data_to_file(self, filename, msg):
        """ Saves data to file asynchronously. This function is called by _process_data_msg() when a buffer is received and saving is enabled.

        Args:
            filename (str): Filename to save data to
            msg (bytestr): Data message received from Bela
        """
        _msg = copy.copy(msg)
        try:
            # make sure there are not two processes writing to the same file
            if filename not in self._saving_file_locks.keys():
                # create lock for file if it does not exist
                self._saving_file_locks[filename] = asyncio.Lock()

            async with self._saving_file_locks[filename]:
                async with aiofiles.open(filename, "a") as f:
                    _json = json.dumps(copy.copy(_msg))
                    await f.write(_json+"\n")

        except Exception as e:
            print_error(f"Error while saving data to file: {e}")

        finally:
            await self._async_remove_item_from_list(self._active_saving_tasks, asyncio.current_task())

    def _generate_filename(self, saving_filename, saving_dir="./"):
        """ Generates a filename for saving data by adding the variable name and a number at the end in case the filename already exists to avoid overwriting saved data. Pattern: varname_filename__idx.ext.  This function is called by start_streaming() and stream_n_values() when saving is enabled. 

        Args:
            saving_filename (str): Root filename

        Returns:
            str: Generated filename
        """

        filename_wo_ext, filename_ext = os.path.splitext(saving_filename)
        # files that follow naming convention, returns list of varname_filename (no __idx.ext)
        matching_files = [os.path.splitext(file)[0].split(
            "__")[0] for file in glob.glob(os.path.join(saving_dir, f"*{filename_wo_ext}*{filename_ext}"))]

        if not matching_files:
            return os.path.join(saving_dir, saving_filename)

        # counts files with the same varname_filename
        idx = max([matching_files.count(item) for item in set(matching_files)])

        return os.path.join(saving_dir, f"{filename_wo_ext}__{idx}{filename_ext}")

    async def _async_remove_item_from_list(self, _list, task):
        """ Removes a task from a list of tasks asynchronously. This function is called by _save_data_to_file() when a task is finished.

        Args:
            _list (list):  A list of tasks
            task (asyncio task): The task to be removed from _list
        """
        _list.remove(task)

    def _check_periods(self, periods, variables):
        """Checks the periods format and values. If periods is an int, it is converted to a list of the same length as variables. If periods is an empty list, it is converted to a list of 1000s. If periods is a list, it is checked that it has the same length as variables and that all values are integers.

        Args:
            periods (_type_): variable periods passed as an argument to a monitoring function
            variables (list): list of variables

        Returns:
            periods (list of ints): Sanity checked periods
        """

        if isinstance(periods, int):
            periods = [periods]*len(variables)
        elif periods == []:
            print_warning("No periods passed, using default value of 1000")
            periods = [1000]*len(variables)

        for period in periods:
            if period < 500 and period > 1:
                warnings.warn(
                    "Periods < 500 will send messages too frequently and may cause the websocket to crash. Use streaming methods instead.")

        assert len(periods) == len(
            variables), "Periods list must have the same length as variables list"
        for p in periods:
            assert isinstance(
                p, int), "Periods must be integers"

        return periods



import asyncio
import nest_asyncio
import websockets
import json
import errno
import struct
import os
from .utils import print_error, print_warning, print_ok


class Watcher:

    def __init__(self, ip="192.168.7.2", port=5555, data_add="gui_data", control_add="gui_control"):
        """ Watcher class

            Args:
                ip (str, optional): Remote address IP. If using internet over USB, the IP won't work, pass "bela.local". Defaults to "192.168.7.2".
                port (int, optional): Remote address port. Defaults to 5555.
                data_add (str, optional): Data endpoint. Defaults to "gui_data".
                control_add (str, optional): Control endpoint. Defaults to "gui_control".
        """

        self.project_name = None
        self._sample_rate = None

        self.ip = ip
        self.port = port
        self.data_add = data_add
        self.control_add = control_add
        self.ws_ctrl_add = f"ws://{self.ip}:{self.port}/{self.control_add}"
        self.ws_data_add = f"ws://{self.ip}:{self.port}/{self.data_add}"
        self.ws_ctrl = None
        self.ws_data = None

        self._list_response_available = asyncio.Event()
        self._list_response = None
        self._log_response_available = asyncio.Event()
        self._log_response = None

        self._watcher_vars = None

        self._mode = "WATCH"

        global _pybela_ws_register
        try:
            _ = _pybela_ws_register
        except NameError:  # initialise _pybela_ws_register only once in runtime
            _pybela_ws_register = {"WATCH": {},
                                   "STREAM":  {},
                                   "LOG":  {},
                                   "MONITOR":  {}}

        self._pybela_ws_register = _pybela_ws_register

        # debug
        self._printall_responses = False

        # event loop needs to be nested - otherwise it conflicts with jupyter's event loop
        nest_asyncio.apply()

    @property
    def sample_rate(self):
        return self._sample_rate

    @property
    def watcher_vars(self):
        """Returns variables in watcher with their properties (name, type, timestamp_mode, log_filename, data_length)

        Returns:
            list of dicts: List of variables in watcher and their properties
        """
        if self._watcher_vars == None:  # populate
            _list = self.list()
            self._watcher_vars = self._filtered_watcher_vars(
                _list["watchers"], lambda var: True)
        return self._watcher_vars   # updates every time start is called

    @property
    def watched_vars(self):
        """Returns a list of the variables in the watcher that are being watched (i.e., whose data is being sent over websockets for streaming or monitoring)

        Returns:
            list of str: List of watched variables
        """
        _list = self.list()
        return self._filtered_watcher_vars(_list["watchers"], lambda var: var["watched"])

    @property
    def unwatched_vars(self):
        """Returns a list of the variables in the watcher that are not being watched (i.e., whose data is NOT being sent over websockets for streaming or monitoring)

        Returns:
            list of str: List of unwatched variables
        """
        _list = self.list()
        return self._filtered_watcher_vars(_list["watchers"], lambda var: not var["watched"])

    # --- public methods --- #

    def connect(self):
        """Attempts to establish a WebSocket connection and prints a message indicating success or failure.

        """
        if self.is_connected():
            return "Already connected"

        async def _async_connect():
            try:
                # Close any open ctrl websocket for the same mode (STREAM, LOG, MONITOR, WATCH)
                if self._pybela_ws_register[self._mode].get(self.ws_ctrl_add) is not None and self._pybela_ws_register[self._mode][self.ws_ctrl_add].open:
                    print_warning(
                        f"pybela doesn't support more than one active connection at a time for a given mode. Closing previous connection for {self._mode} at {self.ws_ctrl_add}.")
                    await self._pybela_ws_register[self._mode][self.ws_ctrl_add].close()

                # Connect to the control websocket
                self.ws_ctrl = await websockets.connect(self.ws_ctrl_add)
                self._pybela_ws_register[self._mode][self.ws_ctrl_add] = self.ws_ctrl

                # Check if the response indicates a successful connection
                response = json.loads(await self.ws_ctrl.recv())
                if "event" in response and response["event"] == "connection":
                    self.project_name = response["projectName"]
                    # Send connection reply to establish the connection
                    self.send_ctrl_msg({"event": "connection-reply"})

                    # Close any open data websocket for the same mode (STREAM, LOG, MONITOR, WATCH)
                    if self._pybela_ws_register[self._mode].get(self.ws_data_add) is not None and self._pybela_ws_register[self._mode][self.ws_data_add].open:
                        print_warning(
                            f"pybela doesn't support more than one active connection at a time for a given mode. Closing previous connection for {self._mode} at {self.data_add}.")
                        await self._pybela_ws_register[self._mode][self.ws_data_add].close()

                    # Connect to the data websocket
                    self.ws_data = await websockets.connect(self.ws_data_add)

                    # Start listener loops
                    self._start_listener(self.ws_ctrl, self.ws_ctrl_add)
                    self._start_listener(self.ws_data, self.ws_data_add)

                    # refresh watcher vars in case new project has been loaded in Bela
                    self._list = self.list()
                    self._sample_rate = self._list["sampleRate"]
                    self._watcher_vars = self._filtered_watcher_vars(self._list["watchers"],
                                                                     lambda var: True)
                    print_ok("Connection successful")
                else:
                    print_error("Connection failed")
            except Exception as e:
                raise ConnectionError(f"Connection failed: {str(e)}.")

        return asyncio.run(_async_connect())

    def stop(self):
        """Closes websockets
        """
        async def _async_stop():
            if self.ws_ctrl is not None and self.ws_ctrl.open:
                await self.ws_ctrl.close()
            if self.ws_data is not None and self.ws_data.open:
                await self.ws_data.close()
        return asyncio.run(_async_stop())

    def is_connected(self):
        return True if (self.ws_ctrl is not None and self.ws_ctrl.open) and (self.ws_data is not None and self.ws_data.open) else False

    def list(self):
        """ Asks the watcher for the list of variables and their properties and returns it
        """
        async def _async_list():
            self.send_ctrl_msg({"watcher": [{"cmd": "list"}]})
            # Wait for the list response to be available
            await self._list_response_available.wait()
            self._list_response_available.clear()  # Reset the event for the next call
            return self._list_response

        return asyncio.run(_async_list())

    def send_ctrl_msg(self, msg):
        """Send control message

        Args:
            msg (str): Message to send to the Bela watcher. Example: {"watcher": [{"cmd": "list"}]}
        """
        self._send_msg(self.ws_ctrl_add, msg)

    # --- private methods --- #

    # start listener

    def _start_listener(self, ws, ws_address):
        """Start listener for messages. The listener is a while True loop that runs in the background and processes messages received in ws as they are received.

        Args:
            ws (websocket): Websocket object
            ws_address (str): Websocket address
        """
        async def _async_start_listener(ws, ws_address):
            try:
                while ws is not None and ws.open:
                    msg = await ws.recv()
                    if self._printall_responses:
                        print(msg)
                    if ws_address == self.ws_data_add:
                        self._process_data_msg(msg)
                    elif ws_address == self.ws_ctrl_add:
                        self._process_ctrl_msg(msg)
                    else:
                        print(msg)
            except Exception as e:
                if ws.open:  # otherwise websocket was closed intentionally
                    handle_connection_exception(
                        ws_address, e, "receiving message")
        asyncio.create_task(
            _async_start_listener(ws, ws_address))

    # send message

    def _send_msg(self, ws_address, msg):
        """Send message to websocket

        Args:
            ws (websocket): Websocket object
            ws_address (str): Websocket address
            msg (str): Message to send
        """
        async def _async_send_msg(ws_address, msg):
            try:
                if ws_address == self.ws_data_add and self.ws_data is not None and self.ws_data.open:
                    await self.ws_data.send(json.dumps(msg))
                elif ws_address == self.ws_ctrl_add and self.ws_ctrl is not None and self.ws_ctrl.open:
                    await self.ws_ctrl.send(json.dumps(msg))
            except Exception as e:
                handle_connection_exception(ws_address, e, "sending message")
                return 0
        asyncio.run(_async_send_msg(ws_address, msg))

    # process messages

    def _process_data_msg(self, msg):  # method overwritten by streamer
        """Process data message. This method is overwritten by the streamer.

        Args:
            msg (str): Bytestring with data
        """
        pass

    def _process_ctrl_msg(self, msg):
        """Process control message

        Args:
            msg (str): Control message to process
        """
        _msg = json.loads(msg)

        if "watcher" in _msg.keys():
            if "logFileName" in _msg["watcher"].keys():  # response to log cmd
                self._log_response = _msg["watcher"]
                self._log_response_available.set()
            elif "sampleRate" in _msg["watcher"].keys():  # response to list cmd
                self._list_response = _msg["watcher"]
                self._list_response_available.set()

    def _parse_binary_data(self, binary_data, timestamp_mode, _type):
        """Binary data parser. This method is used both by the streamer and the logger to parse the binary data buffers.

        Args:
            binary_data (bytestring): String of bytes to parse
            timestamp_mode (str): Timestamp mode ("sparse" or "dense")
            _type (str): Type of the variable ("f", "j", "i", "c", "d")

        Returns:
            dict: Dictionary with parsed buffer and timestamps
        """

        _type = 'i' if _type == 'j' else _type

        data_length = self.get_data_length(_type, timestamp_mode)
        # the format is the same for both logger and streamer so the parsing method is shared

        parsed_buffer = None
        if self._mode == "STREAM" or self._mode == "LOG":
            # sparse mode
            if timestamp_mode == "sparse":
                # ensure that the buffer is the correct size (remove padding)
                binary_data = binary_data[:struct.calcsize("Q") + data_length*struct.calcsize(
                    _type)+data_length*struct.calcsize("I")]

                ref_timestamp, *_buffer = struct.unpack('Q' + f"{_type}" * data_length
                                                        + 'I'*data_length, binary_data)
                data = _buffer[:data_length]
                # remove padding
                rel_timestamps = _buffer[data_length:][:data_length]

                parsed_buffer = {"ref_timestamp": ref_timestamp,
                                 "data": data, "rel_timestamps": rel_timestamps}

            elif timestamp_mode == "dense":
                # ensure that the buffer is the correct size
                # binary_data = binary_data[:struct.calcsize(
                #     'Q')+data_length*struct.calcsize(_type)]

                ref_timestamp, * \
                    data = struct.unpack(
                        'Q' + f"{_type}"*int((len(binary_data) - struct.calcsize('Q'))/struct.calcsize(_type)), binary_data) 

                parsed_buffer = {
                    "ref_timestamp": ref_timestamp, "data": data}

        elif self._mode == "MONITOR":
            ref_timestamp, *_buffer = struct.unpack('Q' + f"{_type}" * int(
                (len(binary_data) - struct.calcsize("Q")) // struct.calcsize(_type)), binary_data)
            # size of the buffer is not fixed as in the other modes

            parsed_buffer = {
                "timestamp": ref_timestamp, "value": _buffer[0]}

        return parsed_buffer

    # --- utils --- #

    def get_latest_timestamp(self):
        return self.list()["timestamp"]

    async def _async_remove_item_from_list(self, _list, task):
        """Remove item from list. Used to remove listeners from the list of listeners.

        Args:
            _list (list): list of tasks
            task (asyncio.Task): task to be removed
        """
        _list.remove(task)

    def _filtered_watcher_vars(self, watchers, filter_func):
        """Filter variables in watcher depending on condition given by filter_func

        Args:
            watchers (list of dicts): List of variables in watcher and their properties
            filter_func (function): filter function

        Returns:
            dict: Filtered variables
        """
        return [{
            "name": var["name"],
            "type": var["type"],
            "timestamp_mode":"sparse" if var["timestampMode"] == 1 else "dense" if var["timestampMode"] == 0 else None,
            # "log_filename": var["logFileName"], # this is updated every time log is called so better not to store it
            "data_length": self.get_data_length(var["type"], "sparse" if var["timestampMode"] == 1 else "dense" if var["timestampMode"] == 0 else None,),
            "monitor": var["monitor"]
        }
            for var in watchers if filter_func(var)]

    def _var_arg_checker(self, variables):
        """ Checks if variables passed to a function are passed as names in a list. It also checks if the variables requested are in the watcher. If none are passed, returns all variables in watcher.

        Args:
            variables (list of str): Variables arg passed to a function
        """

        if len(variables) == 0:
            # if no variables are specified, stream all watcher variables (default)
            return [var["name"] for var in self.watcher_vars]

        variables = variables if isinstance(variables, list) else [
            variables]  # variables should be a list of strings

        # check if variables are in watcher
        for var in variables:
            if var not in [v["name"] for v in self.watcher_vars]:
                raise ValueError(
                    f"Variable {var} is not in the watcher. Please check the list of variables in the watcher with watcher.list().")

        return variables

    def _generate_local_filename(self, local_path):
        # if file already exists, throw a warning and add number at the end of the filename
        new_local_path = local_path  # default
        if os.path.exists(local_path):
            base, ext = os.path.splitext(local_path)
            counter = 1
            new_local_path = local_path
            while os.path.exists(new_local_path):
                new_local_path = f"{base}_{counter}{ext}"
                counter += 1
            print_warning(
                f"{local_path} already exists. Renaming file to {new_local_path}")

        return new_local_path

    def get_prop_of_var(self, var_name, prop):
        """Get property of variable. Properties: name, type, timestamp_mode, log_filename, data_length

        Args:
            var_name (str): Variable name
            prop (str): Requested property

        Returns:
            (type depending on prop): Requested property
        """
        return next(
            (v[prop] for v in self.watcher_vars if v['name'] == var_name), None)

    def get_data_byte_size(self, var_type):
        """Returns the byte size of the data type

        Args:
            var_type (str): Variable type

        Returns:
            int: byte size of the variable type 
        """
        data_byte_size_map = {
            "f": 4,
            "j": 4,
            "i": 4,
            "c": 8,
            "d": 8,
        }
        return data_byte_size_map.get(var_type, 0)

    def get_data_length(self, var_type, timestamp_mode):
        """Data length in the buffer

        Args:
            var_type (str): Variable type
            timestamp_mode (str): Timestamp mode

        Returns:
            int: Data length in the buffer (number of elements) 
        """
        dense_map = {
            "f": 1024,
            "j": 1024,
            "i": 1024,
            "c": 512,
            "d": 512,
        }
        sparse_map = {
            "f": 512,
            "j": 512,
            "i": 512,
            "c": 341,
            "d": 341,
        }
        if timestamp_mode == "dense":
            return dense_map.get(var_type, 0)
        if timestamp_mode == "sparse":
            return sparse_map.get(var_type, 0)

        else:
            # return error message
            return 0

    def get_buffer_size(self, var_type, timestamp_mode):
        """Returns the buffer size in bytes for buffers stored in a log file. This is the size of the buffer that is sent over websockets.

        Args:
            var_type (str): Variable type
            timestamp_mode (str): Timestamp mode

        Returns:
            int: Buffer size in bytes
        """
        # for logging
        data_length = self.get_data_length(var_type, timestamp_mode)
        if timestamp_mode == "sparse":
            if self.get_data_byte_size(var_type) == 4:
                return struct.calcsize('Q')+data_length*struct.calcsize(var_type)+data_length*struct.calcsize("I")
            elif self.get_data_byte_size(var_type) == 8:
                return struct.calcsize('Q')+data_length*struct.calcsize(var_type)+data_length*struct.calcsize("I")+4
        elif timestamp_mode == "dense":
            return struct.calcsize('Q')+data_length*struct.calcsize(var_type)
        else:
            # return error message
            return 0

    # destructor

    def __del__(self):
        self.stop()  # stop websockets


def handle_connection_exception(ws_address, exception, action):
    bela_msg = "Make sure Bela is connected to the same network as your computer, that the IP address is correct, and that there is a project running on Bela."
    print_error(
        f"WebSocket exception while connecting to {ws_address}: {exception}.  {bela_msg}")
    if isinstance(exception, OSError):
        if exception.errno == errno.ECONNREFUSED:
            raise ConnectionError(
                f"Error {exception.errno}: Connection refused while connecting to {ws_address}. {bela_msg}")
        elif exception.errno == errno.ENETUNREACH:
            raise ConnectionError(
                f"Error {exception.errno}: Network is unreachable while connecting to {ws_address}.  {bela_msg}")
        else:
            raise ConnectionError(
                f"Error {exception.errno} while connecting to {ws_address}.  {bela_msg}")
    else:
        print_error(f"Error while {action}: {exception}.  {bela_msg}")



import setuptools
from pathlib import Path

long_description = open(f"{Path(__file__).parent}/readme.md").read()

setuptools.setup(
    name="pybela",
    version="0.0.0.post1",
    author="Teresa Pelinski",
    author_email="teresapelinski@gmail.com",
    description="pybela allows interfacing with Bela, the embedded audio platform, using Python. pybela provides a convenient way to stream, log, and monitor sensor data from Bela to the host machine.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/BelaPlatform/pybela",
    packages=setuptools.find_packages(),
    install_requires=[
        "jupyter",
        "panel",
        "jupyter-bokeh",
        "bitarray",
        "notebook",
        "websockets",
        "bokeh",
        "ipykernel",
        "nest-asyncio",
        "aiofiles",
        "paramiko"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
)




class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def print_error(message):
    print(bcolors.FAIL + message + bcolors.ENDC)


def print_info(message):
    print(bcolors.OKBLUE + message + bcolors.ENDC)


def print_warning(message):
    print(bcolors.WARNING + message + bcolors.ENDC)


def print_ok(message, end='\n', flush=False):
    print(bcolors.OKGREEN + message + bcolors.ENDC, end=end, flush=flush)
